#include <SDL/SDL.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <limits.h>
#include <stdlib.h> /* random() */

/* Generate a high-colour pixel for any SDL surface with */
/* any 16-bit format (555, 565, etc.)                    */

static Uint16 CreateHicolorPixel( SDL_PixelFormat *fmt, Uint8 red, Uint8 green, Uint8 blue ) {
  Uint16 value;
  value = ((red >> fmt->Rloss) << fmt->Rshift ) +
          ((green >> fmt->Gloss) << fmt->Gshift ) +
          ((blue >> fmt->Bloss) << fmt->Bshift );
  return value;
}

/* Mandelbrot stuff */

#define num_pics 280
/*const int num_pics = 1;*/          /* number of pictures to capture */

const int screen_width=700;      /* drawing area width in pixels */
const int screen_height=600;     /* drawing area height in pixels */
const long max_iterations=16384; /* give up point for mandelbrot tests */

const double mandel_limit = 4.0; /* mathematically, 2 squared = 4 */
const double center_r = -0.75;   /* center of the mandelbrot set, real axis */
const double center_i = 0;       /* center of the set, imaginary axis */

struct timeval init_time;        /* time when mandelbrot generation started */

Uint8 *pics[ num_pics ];         /* the pictures */

static Uint16 get_mandel_pixel( SDL_Surface *screen, Uint8 colour ) {
  int red;
  int green;
  int blue;

  red=(colour&7)*32;         /* bits 1,2,3 => 0..8 */
  green=((colour&24)>>3)*32; /* bits 4,5,6 => 0..8 */
  blue=(colour>>6)*64;       /* bits 7,8   => 0..4 */

/*
printf( "%d => %d %d %d => %d\n", colour, red, green, blue,
  CreateHicolorPixel( screen->format, red, green, blue ) );
*/
  return CreateHicolorPixel( screen->format, red, green, blue );

} /* get mandel pixel */


/* Run Length Encoding I/O ----------------------------------------- */
/*                                                                   */
/* A very simple run-length encoding file reader/writer because the  */
/* files get too big without some kind of compression.  RLE reduces  */
/* the picture of the whole Mandelbrot set to about 1/5 of its       */
/* uncompressed size.  Your milage may vary.                         */
/*                                                                   */
/* File format: pairs of a length byte and a character byte (with 1  */
/*   exception for length 0)                                         */
/*   length 0: 8 uncompressed bytes follow                           */
/*   length n: repeat character byte n times                         */
/*                                                                   */
/* Logic:                                                            */
/*   If last two bytes were different, and the next byte is also     */
/*     different, write next 8 bytes uncompressed.                   */
/*   If last byte was the same as this byte, keep counting how many  */
/*     more are the same (up to 255).                                */
/*   If the byte changes, or the maximum count of 255 is reached,    */
/*     write what we've got.                                         */
/*   Always start with 8 uncompressed bytes                          */
/* ----------------------------------------------------------------- */

Uint8 last_mandel_byte=UCHAR_MAX;          /* value of last byte (0 if none) */
Uint8 last_mandel_byte2=UCHAR_MAX;           /* byte before that (0 if none) */
int last_mandel_byte_count=INT_MAX;  /* number of bytes the same (0 if none) */
                                     /* or -1 if counting uncompressed bytes */

/* --- Run length encoding writing --- */

void create_mandel( FILE **f ) {
  *f = fopen( "mandel.dat", "w" );
  fprintf( *f, "%c", 0 );                       /* write uncompressed */
  last_mandel_byte_count=-8;
}

static void flush_mandel( FILE *f ) {
  if ( last_mandel_byte_count > 0 ) {
     fprintf( f, "%c%c", last_mandel_byte_count, last_mandel_byte );
  } else if ( last_mandel_byte_count < 0 ) {
     while( last_mandel_byte_count < 0 ) {
        fprintf( f, "%c", 0 );
        last_mandel_byte_count++;
     }
  }
  last_mandel_byte_count = 0;
}

static void write_mandel( FILE *f, Uint8 byte ) {
  if ( last_mandel_byte_count == 0 ) {               /* nothing so far? */
     last_mandel_byte2 = last_mandel_byte;
     last_mandel_byte = byte;                           /* we got 1 now */
     last_mandel_byte_count=1;
  } else if ( last_mandel_byte_count < 0 ) {     /* doing uncompressed? */
     last_mandel_byte_count++;                            /* count down */
     last_mandel_byte2 = last_mandel_byte;
     last_mandel_byte = byte;
     fprintf( f, "%c", byte );                    /* write uncompressed */
  } else if ( last_mandel_byte != byte ) {
     if ( last_mandel_byte2 != last_mandel_byte && last_mandel_byte_count==1 ) {
        fprintf( f, "%c%c%c", 0, last_mandel_byte, byte );
        last_mandel_byte_count=-6;
     } else {
        flush_mandel( f );                            /* flush what we got */
        last_mandel_byte2 = last_mandel_byte;
        last_mandel_byte = byte;                      /* and we're back */
        last_mandel_byte_count=1;                               /* to 1 */
     }
  } else if ( last_mandel_byte_count == 255 ) {           /* maxed out? */
     flush_mandel( f );                                     /* flush it */
     last_mandel_byte2 = last_mandel_byte;
     last_mandel_byte = byte;                              /* and we're */
     last_mandel_byte_count=1;                             /* back to 1 */
  } else {
     last_mandel_byte_count++;
  }
} /* write mandel */

static void close_mandel( FILE *f ) {
  flush_mandel( f );
  fclose( f );
}

/* --- Run length encoding reading --- */

void open_mandel( FILE **f ) {
  *f = fopen( "mandel.dat", "r" );
  last_mandel_byte_count=0;
}

void next_mandel( FILE *f ) {
/* NOT TESTED */
  int ch;
  while ( last_mandel_byte_count < 0 ) {
    ch = fgetc( f );
    last_mandel_byte_count++;
  }
}

Uint8 read_mandel( FILE *f ) {
int ch;
  if ( last_mandel_byte_count < 0 ) {
     ch = fgetc( f );
     last_mandel_byte_count++;
  } else if ( last_mandel_byte_count > 0 ){
     ch = last_mandel_byte;
     last_mandel_byte_count--;
  } else {
     last_mandel_byte_count = fgetc( f );
     if ( last_mandel_byte_count > 0 ) {
        last_mandel_byte = fgetc( f );
        last_mandel_byte_count--;
        ch = last_mandel_byte;
     } else {
        last_mandel_byte_count=-8;
        ch = read_mandel( f );                /* RECURSION */
     }
  }
  return (char) ch;
}


/* GET MANDEL VALUE ------------------------------------------------ */
/*                                                                   */
/* Evaluatuate how close point (c_z, c_i ) in complex number space   */
/* is to the Mandelbrot set.  Return a number between 0 and          */
/* max_iterations.  A value of max_iterations implies that the point */
/* is probably a member of the Mandelbrot set.                       */
/* ----------------------------------------------------------------- */

static int get_mandel_value( double c_r, double c_i ) {
  int loop_count = 1;
  double z_r     = c_r;
  double z_i     = c_i;
  double z_r2;
  double z_i2;

  do {
     z_i2 = z_i*z_i;
     z_r2 = z_r*z_r;
     z_i = 2.0 * z_r * z_i + c_i;
     z_r = z_r2 - z_i2 + c_r;
     loop_count++;
  } while ( (z_r2 + z_i2 < mandel_limit) && (loop_count != max_iterations) );
  return loop_count;

} /* get mandel value */

/* GET MANDEL VALUE 4 ---------------------------------------------- */
/*                                                                   */
/* Calculate 4 points at once.  C has no function "inlining".  This  */
/* is slightly faster than calling get_mandel_value 4 times.         */
/* ----------------------------------------------------------------- */

static void get_mandel_value4( double point_r, double point_i, double step_size,
  int *count, int *count_b, int *count_c, int *count_d ) {

  int loop_count;
  double c_r;
  double c_i;
  double z_r;
  double z_i;
  double z_r2;
  double z_i2;

  /* top-left */

  loop_count = 0;
  c_r = point_r;
  c_i = point_i;
  z_r = c_r;
  z_i = c_i;
  do {
     z_i2 = z_i*z_i;
     z_r2 = z_r*z_r;
     z_i = 2.0 * z_r * z_i + c_i;
     z_r = z_r2 - z_i2 + c_r;
     loop_count++;
  } while ( (z_r2 + z_i2 < mandel_limit) && (loop_count != max_iterations) );
  *count = loop_count;

  /* top-right */

  loop_count = 0;
  c_r = point_r+step_size;
  c_i = point_i;
  z_r = c_r;
  z_i = c_i;
  do {
     z_i2 = z_i*z_i;
     z_r2 = z_r*z_r;
     z_i = 2.0 * z_r * z_i + c_i;
     z_r = z_r2 - z_i2 + c_r;
     loop_count++;
  } while ( (z_r2 + z_i2 < mandel_limit) && (loop_count != max_iterations) );
  *count_b = loop_count;

  /* bottom-left */

  loop_count = 0;
  c_r = point_r;
  c_i = point_i+step_size;
  z_r = c_r;
  z_i = c_i;
  do {
     z_i2 = z_i*z_i;
     z_r2 = z_r*z_r;
     z_i = 2.0 * z_r * z_i + c_i;
     z_r = z_r2 - z_i2 + c_r;
     loop_count++;
  } while ( (z_r2 + z_i2 < mandel_limit) && (loop_count != max_iterations) );
  *count_c = loop_count;

  /* bottom-right */

  loop_count = 0;
  c_r = point_r+step_size;
  c_i = point_i+step_size;
  z_r = c_r;
  z_i = c_i;
  do {
     z_i2 = z_i*z_i;
     z_r2 = z_r*z_r;
     z_i = 2.0 * z_r * z_i + c_i;
     z_r = z_r2 - z_i2 + c_r;
     loop_count++;
  } while ( (z_r2 + z_i2 < mandel_limit) && (loop_count != max_iterations) );
  *count_d = loop_count;
} /* get_mandel_value 4 */


/* GET AREA COMPLEXITY --------------------------------------------- */
/*                                                                   */
/* Return the percentage complexity of the patterns in an area of    */
/* the Mandelbrot set.  0 is uniform colour, 100 is alternating      */
/* black and white.  This is used to find an "interesting" area to   */
/* view.                                                             */
/* ------------------------------------------------------------------*/

static double get_area_complexity( double point_r, double point_i, double magnification ) {
  int test_width = (double)screen_width*0.1;
  int test_height = (double)screen_height*0.1;
  double step_size = INT_MAX;
  double first_r = INT_MAX;
  double first_i = INT_MAX;
  double c_r = INT_MAX;
  double c_i = INT_MAX;

  int colour;
  int last_colour = INT_MAX;
  double total_diff=0;
  int x_diff=0;
  int y_diff=0;
  int test_x = INT_MAX;
  int test_y = INT_MAX;
  int loop_count = INT_MAX;

  step_size = 0.25 / ((test_width) * magnification);
  first_r = point_r - (test_width/2)*step_size;
  first_i = point_i - (test_height/2)*step_size;

  /* test horizontally */

  c_i = first_i;
  for (test_y=0; test_y<test_height; test_y++) {
    c_r = first_r;
    for (test_x=0; test_x<test_width; test_x++) {

      loop_count = get_mandel_value( c_r, c_i );
      if (loop_count == max_iterations ) {       /* Mandelbrot set */
         colour = 0;                            /* is always black */
      } else {
         colour = loop_count % 256;             /* 16384 => 0..255 */
      }
      if ( test_x == 0 ) {
         last_colour = colour;
      } else if ( last_colour != colour ) {
         x_diff += 1;
      }
      c_r = c_r + step_size;
    }
    c_i = c_i + step_size;
  }
  /* x_diff = x_diff / ( (test_width - 1) * test_height ); */
  x_diff = (100 * x_diff) / ( (test_width - 1) * test_height );

  /* test vertically */

  c_r = first_r;
  for (test_x=0; test_x<test_width; test_x++) {
    c_i = first_i;
    for (test_y=0; test_y<test_height; test_y++) {
      loop_count = get_mandel_value( c_r, c_i );
      if (loop_count == max_iterations ) {       /* Mandelbrot set */
         colour = 0;                            /* is always black */
      } else {
         colour = loop_count % 256;             /* 16384 => 0..255 */
      }
      if ( test_y == 0 ) {
         last_colour = colour;
      } else if ( last_colour != colour ) {
         y_diff += 1;
      }
      c_i = c_i + step_size;
    }
    c_r = c_r + step_size;
  }

  /* y_diff = y_diff / ( test_width * (test_height - 1) ); */
  y_diff = ( 100.0 * y_diff ) / ( test_width * (test_height - 1) );

  /* total_diff=(100.0*( x_diff+y_diff)/2.0 ) / 255.0; */
  total_diff=(x_diff+y_diff)/2.0;

  return total_diff;
} /* get_area_complexity */


/* FIND INTERESTING POINT ------------------------------------------ */
/*                                                                   */
/* Find a point on the current screen with a complex pattern.  Test  */
/* at least 6 points before choosing a point.  The point must have   */
/* at least 5% complexity.                                           */
/* ----------------------------------------------------------------- */

static void find_interesting_point( double start_r, double start_i, double magnification, double *best_r, double *best_i ) {
  const int minimum_loop = 9;                              /* was 10 */
  const double minimum_complexity = 15.0;            /* solid colour */
  const double maximum_complexity = 97.0;        /* pure white noise */
  double step_size = INT_MAX;
  double test_r;
  double test_i;
  double test;
  double best;
  double range_r;
  double range_i;
  double first_r;
  double first_i;
  int count;

  count = 1;
  test  = 0;
  best  = 0;

  step_size = 2.5 / ((screen_width) * magnification);
  range_r = screen_width*step_size;
  range_i = screen_height*step_size;

  first_r = start_r - (screen_width/2)*step_size;
  first_i = start_i - (screen_height/2)*step_size;

  printf( "Searching for a new point in (%lf,%lf)-(%lf,%lf)...\n",
     first_r, first_i, first_r+range_r, first_i+range_i );

  /* First, start with the assumption that we may just want to keep */
  /* moving in the current direction. Use this as our first compare */
  /* point. */

  *best_r = start_r;
  *best_i = start_i;
  test = get_area_complexity( start_r, start_i, magnification );
  if ( test < maximum_complexity ) {
     best = test;
     printf( "Test 0: (%lf,%lf) is %lf%% interesting\n", start_r, start_i, best );
  }

  /* Now test other random points to see if they are interesting */

  for (;;) {
    test_r = first_r + range_r*random()/RAND_MAX;
    test_i = first_i + range_i*random()/RAND_MAX;
    test = get_area_complexity( test_r, test_i, magnification );
    if ( test > best && best < maximum_complexity ) {
       best = test;
       *best_r = test_r;
       *best_i = test_i;
       printf( "Test %d: (%lf,%lf) is %lf%% interesting\n", count, test_r, test_i, test );
    }
    if ( count >= minimum_loop && best > minimum_complexity )
       break;
    count++;
  }
  printf( "Chose: (%lf,%lf) at %lf%%.\n", *best_r, *best_i, best );

} /* find interesting point */


/* STARTUP MANDEL -------------------------------------------------- */
/*                                                                   */
/* Prepare the Mandelbrot generator.  Display a welcome message.     */
/* Allocate necessary memory.                                        */
/* ----------------------------------------------------------------- */

static void startup_mandel( void ) {
  int pic = INT_MAX;
  const long pic_size = screen_width * screen_height;

  gettimeofday( &init_time, NULL );
  srandom( init_time.tv_usec );

  printf( "Mandelbrot Animation\n" );
  printf( "  Pictures: %d pictures...\n", num_pics );
  printf( "  Pic size: %ld bytes\n", pic_size );
  printf( "  Total:    %ld megs\n", (pic_size*num_pics)/1024/1024 );
  printf( "\n" );
  printf( "Use kill -9 %d to stop...\n", getpid() );
  printf( "\n" );
  printf( "Allocating memory...\n" );

  /* allocate memory for pictures */

  for (pic=0;pic<num_pics;pic++) {
      pics[ pic ] = malloc( pic_size );
      if ( pics[ pic ] == NULL ) {
         printf( "Unable to allocate memory for picture %d\n", pic );
         break;
      }
  }

} /* startup_mandel */


/* DRAW MANDEL ----------------------------------------------------- */
/*                                                                   */
/* Draw a section of the Mandelbrot set on the screen, center point  */
/* (point_r, point_i) at the specified magnification.  Store the     */
/* picture in position pic in the pics array.                        */
/* ----------------------------------------------------------------- */

static void draw_mandel( FILE *f, SDL_Surface *screen, double point_r,
double point_i, double magnification, int pic ) {
  int screen_x   = INT_MAX;
  int screen_y   = INT_MAX;
  /* Uint8 colour   = UCHAR_MAX; */
  /* int loop_count = INT_MAX; */
  double c_r     = INT_MAX;
  double c_i     = INT_MAX;
  double first_r = INT_MAX;
  double first_i = INT_MAX;
  double step_size = INT_MAX;
  int offset     = INT_MAX;
  Uint16 pixel_color = USHRT_MAX;
  Uint16 *raw_pixels = NULL;
  long pixel = 0;
  struct timeval start_time;
  struct timeval end_time;
  long minutes = -1;
  long seconds = -1;
  long total_hours = -1;
  long total_minutes = -1;
  long i;

int pt1, pt2, pt3, pt4;
int offset2;

  raw_pixels = (Uint16*) screen->pixels;

  printf( "Picture #%d of %d:\n", pic+1, num_pics );
  printf( "  Location: (%lf,%lf) at zoom x%lf\n", point_r, point_i,
    magnification);
/*
  printf( "  Center Complexity: %lf%%\n",
    get_area_complexity( point_r, point_i, magnification ) );
*/

  gettimeofday( &start_time, NULL );

  step_size = 2.5 / (screen_width * magnification);
  first_r = point_r - (screen_width/2)*step_size;
  first_i = point_i - (screen_height/2)*step_size;

  c_i = first_i;
  for (screen_y=0; screen_y<screen_height; screen_y+=2) {
    c_r = first_r;
    offset = (screen->pitch / 2 * screen_y);
    offset2 = (screen->pitch / 2 * (screen_y + 1 ));
    SDL_LockSurface( screen );
    for (screen_x=0; screen_x<screen_width; screen_x+=2) {

      get_mandel_value4( c_r, c_i, step_size, &pt1, &pt2, &pt3, &pt4 );

      /* top-left */

      if (pt1 == max_iterations ) {
         pixel_color = get_mandel_pixel( screen, 0 );
      } else {
         pixel_color = get_mandel_pixel( screen, pt1 % 256 );
      }
      raw_pixels[offset+screen_x]=pixel_color;
      pics[ pic ][ pixel ]=pt1;

      /* top-right */

      if (pt2 == max_iterations ) {
         pixel_color = get_mandel_pixel( screen, 0 );
      } else {
         pixel_color = get_mandel_pixel( screen, pt2 % 256 );
      }
      raw_pixels[offset+screen_x+1]=pixel_color;
      pics[ pic ][ pixel+1 ]=pt2;

      /* bottom-left */

      if (pt3 == max_iterations ) {
         pixel_color = get_mandel_pixel( screen, 0 );
      } else {
         pixel_color = get_mandel_pixel( screen, pt3 % 256 );
      }
      raw_pixels[offset2+screen_x]=pixel_color;
      pics[ pic ][ pixel+screen_width ]=pt3;

      /* bottom-right */

      if (pt4 == max_iterations ) {
         pixel_color = get_mandel_pixel( screen, 0 );
      } else {
         pixel_color = get_mandel_pixel( screen, pt4 % 256 );
      }
      raw_pixels[offset2+screen_x+1]=pixel_color;
      pics[ pic ][ pixel+screen_width+1 ]=pt4;

/*
      loop_count = get_mandel_value( c_r, c_i );
      if (loop_count == max_iterations ) {
         colour = 0;
      } else {
         colour = loop_count % 256;
      }
      pixel_color = get_mandel_pixel( screen, colour );
      raw_pixels[offset+screen_x]=pixel_color;
*/
/*
      pics[ pic ][ pixel ]=colour;
      write_mandel( f, colour );
      pixel++;
*/
      c_r = c_r + step_size + step_size;
      pixel++;
      pixel++;
    }
    SDL_UnlockSurface( screen );

    /* Show progress: drawing every vertical line is too slow */
    /* and not much to see.  Draw every 4th line instead. */

    if ( (screen_x & 3) == 0 )
       SDL_UpdateRect(screen, 0, 0, 0, 0);

    c_i = c_i + step_size + step_size;
    pixel = pixel + screen_width; /* down two lines */
  }
  SDL_UpdateRect(screen, 0, 0, 0, 0);

  /* write to file */

  for (i=0; i<pixel; i++ )
      write_mandel( f, pics[ pic ][ i ] );

  gettimeofday( &end_time, NULL );

  printf( "  Showing:  (%lf,%lf)-(%lf,%lf)\n", first_r, first_i, c_r, c_i );
  minutes=(end_time.tv_sec - start_time.tv_sec) / 60;
  seconds=(end_time.tv_sec - start_time.tv_sec) % 60;
  total_hours=(end_time.tv_sec - init_time.tv_sec) / 3600;
  total_minutes=((end_time.tv_sec - init_time.tv_sec) % 3600) / 60;
  printf( "  Elapsed:  %ld min %ld sec (%ld hr %ld min total)\n",
    minutes, seconds, total_hours, total_minutes );

} /* draw mandel */


/* MANDEL ---- ----------------------------------------------------- */
/*                                                                   */
/* Run the Mandelbrot generator.                                     */
/* ----------------------------------------------------------------- */

static void mandel( SDL_Surface *screen ) {
  int screen_x   = INT_MAX;
  int screen_y   = INT_MAX;
  Uint8 colour   = UCHAR_MAX;
  int offset     = INT_MAX;
  Uint16 pixel_color = USHRT_MAX;
  Uint16 *raw_pixels = NULL;
  double magnification = 1.0;
  long pixel = 0;
  int pic = INT_MAX;
  double point_r;
  double point_i;
  int zoom_count=INT_MAX;
  int move_count=INT_MAX;
  double move_dx=0;
  double move_dy=0;
  double current_r;
  double current_i;
  FILE *f=NULL;

  pic=0;
  zoom_count=1;
  move_count=0;
  current_r = center_r;
  current_i = center_i;

  create_mandel( &f );

restart:

  if ( move_count > 0 ) {
     current_r = current_r + move_dx;
     current_i = current_i + move_dy;
     move_count--;
     if ( move_count==0 ) {
        zoom_count=6+11.0*(float)random()/RAND_MAX;
        printf( "[Done moving...zooming for %d]\n", zoom_count );
     }
  } else if ( zoom_count == 0 ) {
     find_interesting_point( current_r, current_i, magnification,
       &point_r, &point_i );
     move_dx = (point_r - current_r) / 8.0;
     move_dy = (point_i - current_i) / 8.0;
     current_r = current_r + move_dx;
     current_i = current_i + move_dy;
     move_count=7; /* not counting this move right now! */
  } else {
     zoom_count--;
  }

  draw_mandel( f, screen, current_r, current_i, magnification, pic );

  SDL_Delay( 500 );

  if ( pic < num_pics-1 ) {
     pic++;
     pixel=0;
     magnification = magnification*1.1; /* increase by 10% */
     goto restart;
  }

  printf( "Animating (kill -9 %d to stop)...\n", getpid() );

  close_mandel( f );

  raw_pixels = (Uint16*) screen->pixels;

  for (;;) {
  for (pic=0; pic<num_pics; pic++) {
    SDL_LockSurface( screen );
    pixel=0;
    for (screen_y=0; screen_y<screen_height; screen_y++) {
        offset = (screen->pitch / 2 * screen_y );
        for (screen_x=0; screen_x<screen_width; screen_x++) {
            colour=pics[ pic ][ pixel ];
            pixel_color = get_mandel_pixel( screen, colour );
            pixel++;
            raw_pixels[offset+screen_x]=pixel_color;
        }
    }

    SDL_UnlockSurface( screen );
    SDL_UpdateRect(screen, 0, 0, 0, 0);
    SDL_Delay( 150 ); /* Should really calculate */
  }
  SDL_Delay( 1000 );
  for (pic=num_pics-2; pic>=0; pic--) {
    SDL_LockSurface( screen );
    pixel=0;
    for (screen_y=0; screen_y<screen_height; screen_y++) {
        offset = (screen->pitch / 2 * screen_y );
        for (screen_x=0; screen_x<screen_width; screen_x++) {
            colour=pics[ pic ][ pixel ];
            pixel_color = get_mandel_pixel( screen, colour );
            pixel++;
            raw_pixels[offset + screen_x]=pixel_color;
        }
    }
    SDL_UnlockSurface( screen );
    SDL_UpdateRect(screen, 0, 0, 0, 0);
    SDL_Delay( 150 ); /* Should really calculate */
  }
  SDL_Delay( 2000 );
  }

} /* end mandel */


static void playback( SDL_Surface *screen ) {
  int pic    = INT_MAX;
  int offset = INT_MAX;
  Uint16 *raw_pixels = NULL;
  int screen_x   = INT_MAX;
  int screen_y   = INT_MAX;
  Uint8 colour   = UCHAR_MAX;
  Uint16 pixel_color = USHRT_MAX;
  FILE *f    = NULL;

  raw_pixels = (Uint16*) screen->pixels;
  open_mandel( &f );

  for (pic=num_pics-2; pic>=0; pic--) {
    SDL_LockSurface( screen );
    for (screen_y=0; screen_y<screen_height; screen_y++) {
        offset = (screen->pitch / 2 * screen_y );
        for (screen_x=0; screen_x<screen_width; screen_x++) {
            colour = read_mandel( f );
            pixel_color = get_mandel_pixel( screen, colour );
            raw_pixels[offset + screen_x]=pixel_color;
        }
    }
    SDL_UnlockSurface( screen );
    SDL_UpdateRect(screen, 0, 0, 0, 0);
    SDL_Delay( 150 ); /* Should really calculate */
    next_mandel( f );
  }
  SDL_Delay( 2000 );
  close_mandel( f );

} /* playback */


/* MAIN PROGRAM ---------------------------------------------------- */

int main() {
   int result;
   SDL_Surface *screen;
   Uint16 *raw_pixels;
   int x, y;

   nice( 5 );

   result = SDL_Init( SDL_INIT_VIDEO );
   if ( result != 0 ) {
      printf( "Unable to init SDL: %s\n", SDL_GetError() );
      return 1;
   }

   atexit( SDL_Quit );

   screen = SDL_SetVideoMode( screen_width, screen_height, 16, 0 );
   if ( screen == NULL ) {
      printf( "Unable to set video mode: %s\n", SDL_GetError() );
      return 1;
   }

   SDL_LockSurface( screen );
   raw_pixels = (Uint16*) screen->pixels;
   for (x = 0; x < screen_width; x++ ) {
       for (y = 0; y < screen_height; y++ ) {
          Uint16 pixel_color;
          int offset;
          pixel_color = CreateHicolorPixel( screen->format, x/2, 0, y/2 );
          offset = (screen->pitch / 2 * y + x );
          raw_pixels[offset]=pixel_color;
       }
   }

   SDL_UnlockSurface(screen);
   SDL_UpdateRect(screen, 0, 0, 0, 0);

   startup_mandel();
   /* playback( screen ); */
   mandel( screen );

   SDL_Delay( 4000 );
   return 0;

} /* main */

